from __future__ import annotations

import re

from pydantic import BaseModel

host_container_re = re.compile(r"^(?P<host>\d+):(?P<container>\d+)$")
single_port_re = re.compile(r"^\d+$")


class Ports(BaseModel):
    host: int
    container: int

    @staticmethod
    def from_string(string: str) -> Ports:
        if match := host_container_re.match(string):
            return Ports(host=int(match.group("host")), container=int(match.group("container")))
        elif match := single_port_re.match(string):
            return Ports(host=int(string), container=int(string))
        else:
            raise NotImplementedError(string)

    def same_ports(self) -> bool:
        return self.host == self.container



from __future__ import annotations

from enum import Enum, unique
from io import StringIO
from ipaddress import IPv4Address
from pathlib import Path
from typing import Iterable, Literal, Optional, Union

from dotenv import dotenv_values
from pydantic import Field

from komposer.types.base import CamelCaseImmutableBaseModel
from komposer.types.cli import Context
from komposer.types.ports import Ports
from komposer.utils import to_kubernetes_name

Labels = dict[str, Optional[str]]
Annotations = dict[str, Optional[str]]


@unique
class ImagePullPolicy(Enum):
    IF_NOT_PRESENT = "IfNotPresent"
    ALWAYS = "Always"
    NEVER = "Never"


@unique
class PathType(Enum):
    IMPLEMENTATION_SPECIFIC = "ImplementationSpecific"
    EXACT = "Exact"
    PREFIX = "Prefix"


class UnnamedMetadata(CamelCaseImmutableBaseModel):
    labels: Labels = {}
    annotations: Optional[Annotations] = None


class Metadata(UnnamedMetadata):
    name: str

    @staticmethod
    def labels_from_context(context: Context) -> Labels:
        return {
            "repository": context.repository_name_kubernetes,
            "branch": context.branch_name_kubernetes,
        }

    @staticmethod
    def from_context(context: Context, annotations: Optional[Annotations] = None) -> Metadata:
        return Metadata(labels=Metadata.labels_from_context(context), annotations=annotations)

    @staticmethod
    def from_context_with_suffix(
        context: Context, suffix: str, annotations: Optional[Annotations] = None
    ) -> Metadata:
        name = f"{context.manifest_prefix}-{to_kubernetes_name(suffix)}"
        labels = Metadata.labels_from_context(context)

        return Metadata(name=name, labels=labels, annotations=annotations)

    @staticmethod
    def from_context_with_name(
        context: Context, annotations: Optional[Annotations] = None
    ) -> Metadata:
        return Metadata(
            name=context.manifest_prefix,
            labels=Metadata.labels_from_context(context),
            annotations=annotations,
        )


class Item(CamelCaseImmutableBaseModel):
    api_version: str
    kind: str
    metadata: Metadata


class ConfigMap(Item):
    api_version: Literal["v1"] = "v1"
    kind: Literal["ConfigMap"] = "ConfigMap"
    metadata: Metadata
    data: dict[str, Optional[str]]


class Selector(CamelCaseImmutableBaseModel):
    match_labels: Labels = {}


class HostAlias(CamelCaseImmutableBaseModel):
    ip: IPv4Address
    hostnames: list[str] = []


class ConfigMapKeyRef(CamelCaseImmutableBaseModel):
    key: str
    name: str


class ConfigMapEnvironmentVariable(CamelCaseImmutableBaseModel):
    name: str
    value_from: ConfigMapKeyRef

    @staticmethod
    def from_env_file(
        prefix: str, docker_compose_path: Path, relative_env_file_path: Path
    ) -> Iterable[ConfigMapEnvironmentVariable]:
        env_file = docker_compose_path / relative_env_file_path
        env_vars = dotenv_values(env_file)
        config_map_name = f"{prefix}-{to_kubernetes_name(str(relative_env_file_path))}"
        keys = sorted(env_vars.keys())

        for key in keys:
            yield ConfigMapEnvironmentVariable(
                name=key, valueFrom=ConfigMapKeyRef(key=key, name=config_map_name)
            )


class EnvironmentVariable(CamelCaseImmutableBaseModel):
    name: str
    value: Optional[str]

    @staticmethod
    def from_string(string: str) -> Iterable[EnvironmentVariable]:
        env_vars = dotenv_values(stream=StringIO(string))
        items = sorted(env_vars.items())

        for key, value in items:
            yield EnvironmentVariable(name=key, value=value)


class ContainerPort(CamelCaseImmutableBaseModel):
    container_port: int
    host_port: Optional[int] = None

    @staticmethod
    def from_string(string: str) -> ContainerPort:
        ports = Ports.from_string(string)

        if ports.same_ports():
            return ContainerPort(containerPort=ports.container)

        return ContainerPort(containerPort=ports.container, hostPort=ports.host)


class Container(CamelCaseImmutableBaseModel):
    image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT
    image: str
    name: str
    args: Optional[list[str]] = None
    env: list[Union[EnvironmentVariable, ConfigMapEnvironmentVariable]] = []
    ports: Optional[list[ContainerPort]] = []


class TemplateSpec(CamelCaseImmutableBaseModel):
    service_account_name: Optional[str] = None
    host_aliases: list[HostAlias] = []
    containers: list[Container] = []


class Template(CamelCaseImmutableBaseModel):
    metadata: UnnamedMetadata
    spec: TemplateSpec


class DeploymentSpec(CamelCaseImmutableBaseModel):
    replicas: int = Field(1, ge=1)
    selector: Selector
    template: Template


class Deployment(Item):
    api_version: Literal["apps/v1"] = "apps/v1"
    kind: Literal["Deployment"] = "Deployment"
    metadata: Metadata
    spec: DeploymentSpec


class ServicePort(CamelCaseImmutableBaseModel):
    name: str
    port: int
    target_port: int

    @staticmethod
    def from_string(string: str) -> ServicePort:
        ports = Ports.from_string(string)
        name = to_kubernetes_name(string)

        return ServicePort(name=name, targetPort=ports.container, port=ports.host)


class ServiceSpec(CamelCaseImmutableBaseModel):
    ports: list[ServicePort] = []
    selector: Labels


class Service(Item):
    api_version: Literal["v1"] = "v1"
    kind: Literal["Service"] = "Service"
    metadata: Metadata
    spec: ServiceSpec


class IngressTls(CamelCaseImmutableBaseModel):
    hosts: list[str]
    secret_name: str


class ServiceRefPort(CamelCaseImmutableBaseModel):
    number: int

    @staticmethod
    def from_string(string: str) -> ServiceRefPort:
        ports = Ports.from_string(string)

        return ServiceRefPort(number=ports.host)


class ServiceRef(CamelCaseImmutableBaseModel):
    name: str
    port: ServiceRefPort


class Backend(CamelCaseImmutableBaseModel):
    service: ServiceRef


class HttpPath(CamelCaseImmutableBaseModel):
    path: Path
    path_type: PathType
    backend: Backend


class HttpPaths(CamelCaseImmutableBaseModel):
    paths: list[HttpPath]


class IngressRule(CamelCaseImmutableBaseModel):
    host: str
    http: HttpPaths


class IngressSpec(CamelCaseImmutableBaseModel):
    ingress_class_name: str = "nginx-internal"
    tls: Optional[list[IngressTls]] = None
    rules: list[IngressRule] = []


class Ingress(Item):
    api_version: Literal["networking.k8s.io/v1"] = "networking.k8s.io/v1"
    kind: Literal["Ingress"] = "Ingress"
    metadata: Metadata
    spec: IngressSpec


class List(CamelCaseImmutableBaseModel):
    api_version: Literal["v1"] = "v1"
    kind: Literal["List"] = "List"
    # We should be able to use Field(..., discriminator="kind") here
    items: list[Item] = []



from komposer.types import docker_compose, kubernetes
from komposer.types.cli import Context


def generate_service_ports(ports: list[str]) -> list[kubernetes.ServicePort]:
    return [kubernetes.ServicePort.from_string(port) for port in ports]


def generate_service(
    context: Context, service_name: str, service: docker_compose.Service
) -> kubernetes.Service:
    metadata = kubernetes.Metadata.from_context_with_suffix(context, service_name)
    k8s_service = kubernetes.Service(
        metadata=metadata,
        spec=kubernetes.ServiceSpec(
            ports=generate_service_ports(service.ports), selector=metadata.labels
        ),
    )

    return k8s_service


def generate_services(
    context: Context, services: docker_compose.Services
) -> list[kubernetes.Service]:
    return [
        generate_service(context, service_name, service)
        for service_name, service in services.items()
        if service.ports
    ]



import re

from yaml.parser import ParserError

from komposer.core.config_map import generate_config_maps
from komposer.core.deployment import generate_deployment
from komposer.core.extra_manifest import load_extra_manifest
from komposer.core.ingress import generate_ingress_from_services
from komposer.core.service import generate_services
from komposer.exceptions import (
    ComposePortsNotUniqueError,
    DeploymentAnnotationsInvaliYamlError,
    DeploymentAnnotationsNotAMappingError,
    IngressTlsInvalidYamlError,
    IngressTlsNotAListError,
    InvalidServiceNameError,
)
from komposer.types import docker_compose, kubernetes
from komposer.types.cli import Context
from komposer.types.ports import Ports
from komposer.utils import as_json_object, parse_docker_compose_file

rfc_1123_re = re.compile(r"^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$")


def ensure_service_name_lowercase_RFC_1123(compose: docker_compose.DockerCompose) -> None:
    invalid_service_names = [
        service_name
        for service_name in compose.services.keys()
        if not rfc_1123_re.match(service_name)
    ]

    if invalid_service_names:
        raise InvalidServiceNameError(
            f"Invalid lowercase RFC-4122 service names detected: {invalid_service_names}"
        )


def ensure_unique_ports_on_docker_compose(compose: docker_compose.DockerCompose) -> None:
    non_unique_ports: dict[int, list[str]] = {}

    for service_name, service in compose.services.items():
        for port_str in service.ports:
            port = Ports.from_string(port_str).container

            non_unique_ports.setdefault(port, []).append(service_name)

    non_unique_ports = {
        port: service_names
        for port, service_names in non_unique_ports.items()
        if len(service_names) > 1
    }

    if non_unique_ports:
        raise ComposePortsNotUniqueError(f"Non-unique ports detected: {non_unique_ports}")


def ensure_ingress_tls_is_valid_yaml(context: Context) -> None:
    try:
        ingress_tls = context.ingress.tls
    except ParserError:
        raise IngressTlsInvalidYamlError("The Ingress TLS value is not a valid YAML")

    if ingress_tls is None:
        return

    if not isinstance(ingress_tls, list):
        raise IngressTlsNotAListError("The Ingress TLS value is not a list")


def ensure_deployment_annotations_is_valid_yaml(context: Context) -> None:
    try:
        deployment_annotations = context.deployment.annotations
    except ParserError:
        raise DeploymentAnnotationsInvaliYamlError(
            "The Deployment annotations value is not a valid YAML"
        )

    if deployment_annotations is None:
        return

    if not isinstance(deployment_annotations, dict):
        raise DeploymentAnnotationsNotAMappingError(
            "The Deployment annotations value is not a mapping"
        )


def generate_manifest_from_docker_compose(context: Context) -> dict:
    # Parse docker compose file
    compose = parse_docker_compose_file(context.docker_compose_path)

    # Ensures Docker Compose is supported
    ensure_deployment_annotations_is_valid_yaml(context)
    ensure_ingress_tls_is_valid_yaml(context)
    ensure_unique_ports_on_docker_compose(compose)
    ensure_service_name_lowercase_RFC_1123(compose)

    # Generate configmaps
    config_maps = generate_config_maps(context, compose)

    # Generate pod
    deployment = generate_deployment(context, compose.services)

    # Generate services
    services = generate_services(context, compose.services)

    # Generate ingress
    ingress = generate_ingress_from_services(context, compose.services)

    # Load external manifest
    extra_manifest = load_extra_manifest(context)

    # Return manifest
    items = [*config_maps, deployment, *services]

    if ingress:
        items.append(ingress)

    manifest = kubernetes.List(items=items)

    # Convert to JSON object so that we can append the extra manifest as is
    manifest_dict = as_json_object(manifest)
    manifest_dict["items"].extend(extra_manifest)

    return manifest_dict



import itertools
from typing import Literal

import yaml
from pydantic import parse_obj_as

from komposer.types import kubernetes
from komposer.types.base import CamelCaseImmutableBaseModel
from komposer.types.cli import Context


class ExtraManifest(CamelCaseImmutableBaseModel):
    api_version: Literal["v1"] = "v1"
    kind: Literal["List"] = "List"
    items: list[kubernetes.Item] = []


def update_item_metadata_labels(item: dict, labels: dict) -> None:
    item["metadata"].setdefault("labels", {}).update(labels)


def update_item_metadata_name(item: dict, manifest_prefix: str) -> None:
    item["metadata"]["name"] = f"{manifest_prefix}-{item['metadata']['name']}"


def update_item_env_configmapkeyref_name(item: dict, manifest_prefix: str) -> None:
    containers = item.get("spec", {}).get("template", {}).get("spec", {}).get("containers", [])
    envs = itertools.chain(*(container.get("env", []) for container in containers))
    config_map_key_refs = (env.get("valueFrom", {}).get("configMapKeyRef", {}) for env in envs)

    for config_map_key_ref in config_map_key_refs:
        config_map_key_ref["name"] = f"{manifest_prefix}-{config_map_key_ref['name']}"


def load_extra_manifest(context: Context) -> list[dict]:
    # Skip if no extra manifest
    if context.extra_manifest_path is None:
        return []

    # Parse manifest
    extra_manifest_raw = yaml.safe_load(context.extra_manifest_path.open())

    # just validate that the format of the file
    parse_obj_as(ExtraManifest, extra_manifest_raw)

    # this is necessary because we need to manipulate the original item
    # and the item can be undefined in our list of types
    extra_manifest_items: list[dict] = extra_manifest_raw.get("items", [])

    # Update metadate labels and names and env's configmap names
    labels = kubernetes.Metadata.labels_from_context(context)
    manifest_prefix = context.manifest_prefix

    for item in extra_manifest_items:
        update_item_metadata_labels(item, labels)
        update_item_metadata_name(item, manifest_prefix)
        update_item_env_configmapkeyref_name(item, manifest_prefix)

    return extra_manifest_items



from pathlib import Path
from typing import Any, Optional

import yaml

from komposer.types.base import ImmutableBaseModel
from komposer.utils import to_kubernetes_name


def parse_str_as_yaml(value_str: Optional[str]) -> Any:
    if value_str is None:
        return None

    value = yaml.safe_load(value_str)

    return value


class DeploymentContext(ImmutableBaseModel):
    annotations_str: Optional[str] = None
    service_account_name: Optional[str] = None

    @property
    def annotations(self) -> Optional[Any]:
        return parse_str_as_yaml(self.annotations_str)


class IngressContext(ImmutableBaseModel):
    tls_str: Optional[str] = None

    @property
    def tls(self) -> Optional[Any]:
        return parse_str_as_yaml(self.tls_str)


class Context(ImmutableBaseModel):
    docker_compose_path: Path
    project_name: str
    branch_name: str
    repository_name: str
    default_image: str
    ingress_for_service: Optional[str] = None
    extra_manifest_path: Optional[Path] = None
    deployment: DeploymentContext
    ingress: IngressContext

    @property
    def project_name_kubernetes(self) -> str:
        return to_kubernetes_name(self.project_name)

    @property
    def branch_name_kubernetes(self) -> str:
        return to_kubernetes_name(self.branch_name)

    @property
    def repository_name_kubernetes(self) -> str:
        return to_kubernetes_name(self.repository_name)

    @property
    def manifest_prefix(self) -> str:
        return "-".join(
            [
                self.project_name_kubernetes,
                self.repository_name_kubernetes,
                self.branch_name_kubernetes,
            ]
        )



class KomposerException(Exception):
    pass


class IngressTlsException(KomposerException):
    pass


class DeploymentAnnotationsException(KomposerException):
    pass


class ServiceNotFoundError(KomposerException):
    pass


class ComposePortsNotUniqueError(KomposerException):
    pass


class InvalidServiceNameError(KomposerException):
    pass


class IngressTlsInvalidYamlError(IngressTlsException):
    pass


class IngressTlsNotAListError(IngressTlsException):
    pass


class DeploymentAnnotationsInvaliYamlError(DeploymentAnnotationsException):
    pass


class DeploymentAnnotationsNotAMappingError(DeploymentAnnotationsException):
    pass



import json
import shlex
from pathlib import Path
from typing import Any, Optional, Union, cast

import stringcase
import yaml
from dotenv import dotenv_values
from pydantic import BaseModel

from komposer.types import docker_compose


def to_kubernetes_name(string: str) -> str:
    name: str = stringcase.spinalcase(string.strip().lower().replace("/", "-").replace(":", "-"))

    if name.startswith("-"):
        name = name[1:]

    return name


def parse_docker_compose_file(compose_path: Path) -> docker_compose.DockerCompose:
    with compose_path.open() as f:
        content = yaml.load(f, Loader=yaml.Loader)

    config = docker_compose.DockerCompose.parse_obj(content)

    return config


def parse_env_file(env_file: Path) -> dict[str, Optional[str]]:
    environment = dotenv_values(env_file)

    return environment


def command_to_args(command: Optional[Union[str, list[str]]]) -> Optional[list[str]]:
    if isinstance(command, str):
        return shlex.split(command)

    return command


def as_json_object(type_: BaseModel) -> dict[str, Any]:
    return cast(dict[str, Any], json.loads(type_.json(by_alias=True)))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['komposer', 'komposer.core', 'komposer.types']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'click>=8.1.3,<9.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'stringcase>=1.2.0,<2.0.0']

entry_points = \
{'console_scripts': ['komposer = komposer.cli:main']}

setup_kwargs = {
    'name': 'komposer',
    'version': '0.1.0.dev8',
    'description': 'Tool to convert a Docker Compose file into a Kubernetes manifest',
    'long_description': '# komposer\n\n## To-do\n\n- disallow port mapping form docker compose\n- set ingress annotations from CLI as file\n- set deployment annotations from CLI as file\n- accept repository and branch name only as kubernetes names\n- set ingress paths from CLI\n- needs the ingress class name, mandatory if ingress is selected\n- set custom resource limits\n',
    'author': 'Daniele Esposti',
    'author_email': 'daniele.esposti@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/expobrain/komposer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Optional

from komposer.exceptions import ServiceNotFoundError
from komposer.types import docker_compose, kubernetes
from komposer.types.cli import Context
from komposer.utils import to_kubernetes_name

DEFAULT_INGRESS_ANNOTATIONS: kubernetes.Annotations = {
    "cert-manager.io/cluster-issuer": "letsencrypt-prod"
}
DEFAULT_CLUSTER_NAME = "svc.cluster.local"


def generate_ingress_path(
    context: Context, service_name: str, service_port: str
) -> kubernetes.HttpPath:
    service_ref_port = kubernetes.ServiceRefPort.from_string(service_port)
    service_name_kubernetes = to_kubernetes_name(service_name)

    http_path = kubernetes.HttpPath(
        path="/",
        pathType=kubernetes.PathType.PREFIX,
        backend=kubernetes.Backend(
            service=kubernetes.ServiceRef(
                name=f"{context.manifest_prefix}-{service_name_kubernetes}",
                port=service_ref_port,
            )
        ),
    )

    return http_path


def generate_ingress_host(context: Context) -> str:
    if context.ingress_for_service is None:
        raise ValueError("Context.ingress_for_service is None")

    return ".".join(
        [
            to_kubernetes_name(context.ingress_for_service),
            context.manifest_prefix,
            DEFAULT_CLUSTER_NAME,
        ]
    )


def generate_ingress_from_services(
    context: Context, services: docker_compose.Services
) -> Optional[kubernetes.Ingress]:
    if context.ingress_for_service is None:
        return None

    if context.ingress_for_service not in services:
        raise ServiceNotFoundError(f"Service {context.ingress_for_service} not found")

    service = services[context.ingress_for_service]
    metadata = kubernetes.Metadata.from_context_with_suffix(
        context, context.ingress_for_service, annotations=DEFAULT_INGRESS_ANNOTATIONS
    )
    host = generate_ingress_host(context)
    paths = [
        generate_ingress_path(context, context.ingress_for_service, port) for port in service.ports
    ]

    ingress = kubernetes.Ingress(
        metadata=metadata,
        spec=kubernetes.IngressSpec(
            rules=[kubernetes.IngressRule(host=host, http=kubernetes.HttpPaths(paths=paths))],
            tls=context.ingress.tls,
        ),
    )

    return ingress
